/* vcwd.c - Virtual current working directory 
 (c) 1995 James Lowe, Jr.

THIS PROGRAM MAY BE DISTRIBUTED UNDER THE TERMS AND CONDITIONS OF THE 
  Free Software Foundation's General Public License (GPL Version 2
  or later)
*/

#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>

/*--------------------------------------------------------------------------*/
static int fmf_linkbacktrace (char * curdirectory , char *name, char * path);
static char * str_rindex ( char * source , char * delim , int rindex, char * dst, char * tmp);
static int absolute_pathname (char * string);
static int num_of_components ( char * str , char * ifs);

/*--------------------------------------------------------------------------*/


#define LINELEN 4000
#define MAXPATHLEN 4000
int main ()
{
  char line [LINELEN], curdirectory [LINELEN], name[LINELEN], path[LINELEN];

 strcpy (curdirectory, "/");

 while (fgets (line, LINELEN - 1, stdin) != (char *) (NULL))
    {
     strtok (line, "\n\r");
     strcpy (name , line );
     fmf_linkbacktrace (curdirectory, name, path);
     puts (curdirectory);

    }

exit (0);
}


int fmf_linkbacktrace (char * curdirectory , char *name, char * path)
{
  char *p, *tmp;
  static int internal_error;
  int i, aa=0, bb=0;

/*    path : is the outgoing, final, to be set, path
      name : is the incoming proposed change */

		/* set path to a sane value in case of error (bailout) */
   strcpy ( path, curdirectory );

   if (internal_error)
     {
       strcpy ( path, name);
       return 0;
     }

   if ( absolute_pathname (name))
    {
        strcpy ( path, name );
        strcpy ( curdirectory, path );
        return 0;
    }

   /* do some simple checks on name */
   if ( ! strncmp (name , "/.", 2 ))
     return -1;

   if ( ! strcmp (name , "."))
     return 0;


   tmp = (char *) malloc ( (MAXPATHLEN + 1) * sizeof (char));
   if ( ! tmp )
     return -1;
   strcpy (tmp , name);

   /* check for leading .. only */
   /* this is supposed to check for embedded '..'; this code
      only handles leading '..'s
   */
   i=0;
   while ((p=strtok (i ? (char *)(NULL) : tmp, "/")) != (char *)(NULL) )
    {
       i++;
       if ( !strcmp (p, "..") )
         {
           if (bb) {
              free (tmp);
              return -1;
            }
           aa++;
         }
       else
         bb++;
    }

  /* remove leading `./'  just to eliminate an exception */
  if ( ! strncmp (name, "./" , 2) )
    memmove ( name, name + 2, strlen (name) -1 );

  /* check for null string in name. */
  /* possible if name was './' */ 

  if ( ! strlen (name))
    {
      return 0;
    }
  
   if ( name[ strlen (name)  - 1 ] == '/' )
      name[ strlen (name) - 1 ] = '\0';

   if ( strstr ( name, "//")  || absolute_pathname (name))
    {
       internal_error++;
       free (tmp);
       return -1;
    }

  /* curdirectory will be altered by this function 
     according to the value of 'name' */
  if ( make_symbolicabsolutepath (curdirectory, name, tmp, (int)1))
   { 
     internal_error ++; 
     free (tmp);
     return -1;
   }

 strcpy ( path, curdirectory);
 free (tmp);
 return 0;
}

static char * str_rindex ( char * source , char * delim , int rindex, char * dst, char * tmp)
{
  char *p , *ret, *ps;
  int i=0, num=0, ix=0;

  if ( ! source )
    return (char *) (NULL);
  strcpy (tmp , source);
  num= num_of_components ( tmp , delim);
  if ( ! num )
    return source + strlen (source);
  ix = num - rindex;
  if (ix < 0 )
    ix=0;

  ps= (char*)(NULL);
  while ( ((p=strtok (i ? (char *)(NULL) : tmp, delim)) != (char *)(NULL)) )
    {
       if ( i == ix )
         {
            ps = p;
            break;
         }
       i++;
    }
  if ( !ps )
    ps = tmp + strlen (source); 

  if ( dst )
    {
      strcpy (dst, ps);
      ret=dst;
    }
  else
   {
     ret= source + (int)( ps - tmp );
   }
 return ret;
}


int make_symbolicabsolutepath (char * curdir, char *usercmd, char * tmp, int dironly)
{
  char *p, *p1, oldstr [MAXPATHLEN];
  int i=0, aa=0 ,  num;

		/* form new symbolic absolute directory by modifying
		   the old 'curdir' based on the 'usercmd'.
		   i.e. if usercmd is "../.." then shorten curdir by
		   2 components and append the part of 'usercmd' that 
		   remains after removing the "../.."
		   */

  strcpy (oldstr, curdir);
  strcpy (tmp , usercmd);

		/* find out how many parents we must retrieve */
		/* by counting '..' in the string. We will assume */
		/* that that they are all leading */

  while ((p=strtok (i ? (char *)(NULL) : tmp, "/")) != (char *)(NULL) )
    {
       if ( !strcmp (p, "..") )
         aa++;
       i++;
    }


		/* check to see that  user is not trying to go past the root */
  strcpy (tmp , curdir);
  num=num_of_components ( tmp , "/");
  if ( num < aa  && aa )
     { return 0; }


		/* if there are leading '..' then obtain pointer in oldstr
		   that is 'aa' directory components from the end of the string.
		   (str_rindex counts in reverse ,aa times, delimited fields and
		   returns pointer)
		   */
   p1 = str_rindex (oldstr , "/", aa , (char*)(NULL), tmp);
     if ( ! p1 )
       { return -1; }
   *p1='\0';    /* truncates oldstr at the right place */

		/* count down from the left end of 'usercmd' a number of times
		   equal to the number of '..' inorder to reveal the named directory
		   components 
		   */ 
   strcpy (tmp , usercmd); i=0;
   p = usercmd;
   while ( ((p=strtok (i ? (char *)(NULL) : tmp, "/")) != (char *)(NULL) ) && i < aa )
    {
       i++;
    }
   if ( i < aa )
      {  return -1; }

		/* now make p point to location in usercmd, since tmp is contaminated
		   with nulls and handle the case of a null p
		*/
   if (p == (char *)(NULL))
     p = usercmd + (int)strlen (usercmd);
   else
     p = usercmd + (int)(p - tmp);

		/* check total length
		 */
   if ( strlen (oldstr) + ( p ? strlen (p) : 0 ) + 2 > MAXPATHLEN )
    { return -1; }

		/* now append the strings oldstr and p 
		*/ 
   strcat ( oldstr, "/");
   if ( p && *p)
     {
       strcat (oldstr, p);
     }

		/* squash double slashes */
   while ( ( p1 = strstr ( oldstr, "//" )) != (char *)(NULL) )
     memmove ( p1, p1 + 1, strlen (p1) );

		/* remove trailing '/'  */
  if ( (oldstr[strlen(oldstr) - 1] == '/' ) && strlen (oldstr) > 1 )
     oldstr [ strlen (oldstr) - 1 ] = '\0';

  if ( dironly )	/* if dironly, then don't do it if oldstr is not a
  			   'stat-able' dir */
    {
        strcpy (curdir, oldstr);
    }
  else 
    {
       strcpy (curdir, oldstr);
    }
  return 0;

}

/* Return 1 if STRING contains an absolute pathname, else 0. */
static int
absolute_pathname (string)
     char *string;
{
  if (!string || !*string)
    return (0);

  if (*string == '/')
    return (1);

/*  if (*(string++) == '.')
    {
      if ((!(*string)) || *string == '/')
	return (1);

      if (*(string++) == '.')
	if (!*string || *string == '/')
	  return (1);
    }
  */
  return (0);
}


static int num_of_components ( char * str , char * ifs)
{
  int i=0;
  char * tmp; 

  tmp = (char *) malloc ( (strlen (str) + 1) * sizeof (char));
  if ( ! tmp )
    return -1;

  strcpy (tmp , str);
  while (strtok (i ? (char *)(NULL) : tmp, ifs))
       i++;
  free (tmp);
  return i;
}

